#!/usr/bin/env python

from __future__ import division, absolute_import, print_function

__author__ = "Marek Rudnicki"

import cPickle as pickle
import hashlib
import os
import time
import datetime
import numpy as np
import sys
import logging
import inspect
import socket
import subprocess
import multiprocessing
import shutil
import tempfile
import string
import imp
import functools

logger = logging.getLogger(__name__)


class _FuncWrap(object):
    def __init__(self, func):
        self.func = func

    def __call__(self, data):
        result = _apply_data(self.func, data)
        return result



def _func_wrap(func):
    @functools.wraps(func)
    def wrap(data):
        result = _apply_data(func, data)
        return result

    return wrap



def _apply_data(func, data):

    start = time.time()
    ans = func(**data)
    dt = time.time() - start

    return ans,dt




def _pkl_name(obj, func, cachedir):
    src = inspect.getsource(func)

    pkl = pickle.dumps((obj, src), -1)
    h = hashlib.sha1(pkl).hexdigest()

    pkl_name = os.path.join(
        cachedir,
        h + '.pkl'
    )

    return pkl_name


def _load_cache(fname):
    logger.info("Loading cache from {}".format(fname))
    with open(fname, 'rb') as f:
        data = pickle.load(f)
    return data


def _dump_cache(obj, fname):
    dirname = os.path.dirname(fname)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    logger.info("Dumping cache to {}".format(fname))

    tmp_fname = fname + ".tmp"
    with open(tmp_fname, 'wb') as f:
        pickle.dump(obj, f, -1)
    os.rename(tmp_fname, fname)




def _serial_map(func, iterable, cfg):

    wrap = _FuncWrap(func)

    for args in iterable:
        result = wrap(args)
        yield result



def _isolated_serial_map(func, iterable, cfg):


    for args in iterable:
        dirname = tempfile.mkdtemp()
        fname = os.path.join(
            dirname,
            'mar_maps_socket'
        )
        p = subprocess.Popen(
            ['python', '-m', 'thorns.run_func', fname]
        )

        module_name = inspect.getfile(func)
        func_name = func.func_name
        data = (module_name, func_name, args)

        ### make a socket
        s = socket.socket(socket.AF_UNIX)
        s.bind(fname)
        s.listen(1)
        conn, addr = s.accept()


        ### send the function and data to the child
        f = conn.makefile('wb')
        pickle.dump(data, f, -1)
        f.close()


        ### receive results from the child
        f = conn.makefile('rb')
        result = pickle.load(f)
        f.close()


        ### cleanup
        conn.close()
        s.close()
        shutil.rmtree(dirname)


        yield result




def _multiprocessing_map(func, iterable, cfg):

    wrap = _FuncWrap(func)

    pool = multiprocessing.Pool()


    results = []
    for args in iterable:
        results.append( pool.apply_async(wrap, (args,)) )


    for result in results:
        yield result.get()




def _playdoh_map(func, iterable, cfg):

    import playdoh

    wrap = _func_wrap(func)

    jobrun = playdoh.map_async(
        wrap,
        iterable,
        machines=cfg['machines'],
        codedependencies=cfg['dependencies'],
    )
    results = jobrun.get_results()

    for result in results:
        yield result



def _ipython_map(func, iterable, cfg):

    from IPython.parallel import Client

    rc = Client()
    rc[:].clear()



    ### Make modules for all dependencies on the engines
    for dep in cfg['dependencies']:
        mod_name = os.path.splitext(
            os.path.basename(dep)
        )[0]


        with open(dep) as f:
            code = f.read()

        # TODO: fix the bug with "\n" in the source code.  Trying to
        # escape the backslash, but does not work here.  Working test
        # file in projects/python_demos/escaping
        # code = code.replace("\\", "\\\\")
        # code = code.replace("\'", "\\\'")
        # code = code.replace("\"", "\\\"")
        code = code.encode('string_escape')

        rc[:].execute(
"""
import imp
import sys

_mod = imp.new_module('{mod_name}')
sys.modules['{mod_name}'] = _mod

exec '''{code}''' in _mod.__dict__

del _mod
""".format(code=code, mod_name=mod_name),
            block=True
        )




    ### Make sure all definitions surrounding the func are present on
    ### the engines (evaluate the code from the file of the func)
    fname = inspect.getfile(func)
    with open(fname) as f:
        code = f.read()


    logger.info("IPython engine IDs: {}".format(rc.ids))


    ## Need to escape all ' and " in order to embed the code into
    ## execute string
    # code = code.replace("\'", "\\\'")
    # code = code.replace("\"", "\\\"")

    code = code.encode('string_escape')


    ## The trick with `exec in {}' is done because we want to avoid
    ## executing `__main__'
    rc[:].execute(
"""
_tmp_dict = dict()
exec '''{code}''' in _tmp_dict
globals().update(_tmp_dict)
del _tmp_dict
""".format(code=code),
        block=True
    )
    # status.wait()

    # res = rc[:].apply(dir)
    # print(res.get())

    wrap = _FuncWrap(func)
    pool = rc.load_balanced_view()

    results = []
    for args in iterable:
        results.append( pool.apply_async(wrap, args) )


    for result in results:
        yield result.get()




def _publish_status(status, where='stdout', func_name=""):


    name = os.path.splitext(
        os.path.basename(sys.argv[0])
    )[0]

    ### Bar
    bar_len = 20

    bar = (
        "O" * int(round(bar_len * status['loaded']/status['all'])) +
        "#" * int(round(bar_len * status['processed']/status['all']))
    )
    bar += "." * (bar_len - len(bar))

    seconds = time.time() - status['start_time']

    msg = "{func_name:<22} [{bar}]  {loaded}/{processed}/{remaining}  {time}".format(
        loaded=status['loaded'],
        processed=status['processed'],
        remaining=(status['all']-status['loaded']-status['processed']),
        bar=bar,
        time=datetime.timedelta(seconds=seconds),
        func_name=func_name
    )



    if where == 'file':
        dirname = status['dir']
        fname = os.path.join(dirname, 'status_' + name)

        if not os.path.exists(dirname):
            os.makedirs(dirname)

        with open(fname, 'w') as f:
            f.write(msg)


    elif where == 'stdout':
        print(msg)






def _get_options(backend, cache, dependencies):

    cfg = {}

    if 'MRmap' in os.environ:
        cfg['backend'] = os.environ['MRmap']
    elif backend is not None:
        cfg['backend'] = backend
    else:
        cfg['backend'] = 'serial'


    if 'MRmachines' in os.environ:
        # TODO: must be parsed
        cfg['machines'] = os.environ['MRmachines']
    else:
        cfg['machines'] = []


    if 'MRdependencies' in os.environ:
        # TODO: must be parsed
        cfg['dependencies'] = os.environ['MRdependencies']
    elif dependencies is not None:
        cfg['dependencies'] = dependencies
    else:
        cfg['dependencies'] = []


    if 'MRcache' in os.environ:
        cfg['cache'] = os.environ['MRcache']
    elif cache is not None:
        cfg['cache'] = cache
    else:
        cfg['cache'] = 'yes'


    return cfg






def apply(func, workdir='work', **kwargs):

    results = map(func, [kwargs], workdir=workdir)
    result = list(results)[0]

    return result




def map(
        func,
        iterable,
        backend='serial',
        cache='yes',
        workdir='work',
        dependencies=None,
        kwargs=None
):
    """Apply func to every item of iterable and return a list of the
    results.  This map supports multiple backends, e.g. 'serial',
    'multiprocessing', 'ipcluster'.

    Parameters
    ----------
    func : function
        The function to be applied to the data
    iterable : list of dicts
        Each dict is applied to the func. The keys of the dicts should
        correspond to the parameters of the func.
    cache : bool or {'yes', 'no', 'redo'}
        If True, each result is loaded instead calculated again.

    TODO: finish the docstring

    """
    cfg = _get_options(
        backend=backend,
        cache=cache,
        dependencies=dependencies
    )

    status = {
        'all':0,
        'loaded':0,
        'processed':0,
        'bar': [],
        'times':[],
        'start_time':time.time(),
        'dir': workdir
    }

    cachedir = os.path.join(workdir, 'map_cache')

    cache_files = []
    hows = []
    todos = []
    for args in iterable:
        args = dict(args)
        if kwargs is not None:
            args.update(kwargs)

        fname = _pkl_name(args, func, cachedir)
        cache_files.append(fname)

        status['all'] += 1

        if (cfg['cache'] == 'yes') and os.path.exists(fname):
            hows.append('load')
        else:
            hows.append('process')
            todos.append(args)


    if cfg['backend'] == 'serial':
        results = _serial_map(func, todos, cfg)
    elif cfg['backend'] in ('multiprocessing', 'm'):
        results = _multiprocessing_map(func, todos, cfg)
    elif cfg['backend'] == 'playdoh':
        results = _playdoh_map(func, todos, cfg)
    elif cfg['backend'] in ('ipython', 'ipcluster'):
        results = _ipython_map(func, todos, cfg)
    elif cfg['backend'] == 'serial_isolated':
        results = _isolated_serial_map(func, todos, cfg)
    else:
        raise RuntimeError("Unknown map() backend: {}".format(cfg['backend']))


    answers = []
    for how,fname in zip(hows,cache_files):

        _publish_status(status, 'file', func_name=func.func_name)
        if how == 'load':
            result = _load_cache(fname)
            status['loaded'] += 1

        elif how == 'process':
            result = next(results)
            status['processed'] += 1

            if cfg['cache'] in ('yes', 'refresh', 'redo'):
                _dump_cache(result, fname)

        else:
            raise RuntimeError("Should never reach this point.")

        ans,dt = result
        status['times'].append(dt)

        answers.append(ans)

    _publish_status(status, 'file', func_name=func.func_name)
    _publish_status(status, 'stdout', func_name=func.func_name)

    return(answers)



"""Utilities.

"""

from thorns.util.dumpdb import (
    dumpdb,
    loaddb,
    get_store,
)

from thorns.util.maps import (
    map,
    apply,
)


def mkdir(dirname):
    if not os.path.exists(dirname):
        os.makedirs(dirname)



"""
Spike analysis software.

"""

from __future__ import division, print_function, absolute_import

__author__ = "Marek Rudnicki"


import os
import sys


if 'THpdb' in os.environ:
    import pdb, sys, traceback
    def info(type, value, tb):
        traceback.print_exception(type, value, tb)
        pdb.pm()
    sys.excepthook = info


if 'THlog' in os.environ:
    import logging

    if os.environ['THlog'] in ('d', 'debug'):
        level = 'DEBUG'

    else:
        level = os.environ['THlog'].upper()

    logger = logging.getLogger()
    logger.setLevel(level)



import thorns.util


from thorns.spikes import (
    make_trains,

    trains_to_array,

    accumulate,
    accumulate_spikes,
    accumulate_trains,
    accumulate_spike_trains,

    select_trains,
    select,
    sel,

    trim_spike_trains,
    trim_trains,
    trim,

    fold_spike_trains,
    fold_trains,
    fold,
)


from thorns.stats import (
    get_duration,

    shuffled_autocorrelogram,
    sac,

    correlation_index,
    ci,

    firing_rate,
    rate,

    psth,

    isih,

    entrainment,

    synchronization_index,
    si,

    count_spikes,
    count,

    period_histogram
)

from thorns.plotting import (
    plot_raster,
    plot_psth,
    plot_neurogram,
    plot_sac,
    plot_period_histogram,
    plot_signal,
    show
)


def gcf():
    import matplotlib.pyplot as plt
    fig = plt.gcf()
    return fig



#!/usr/bin/env python

from setuptools import setup, find_packages

with open('README.rst') as file:
    long_description = file.read()

setup(
    name = "thorns",
    version = "0.1",
    author = "Marek Rudnicki",
    author_email = "marek.rudnicki@tum.de",

    description = "Spike analysis software",
    license = "GPLv3",
    url = "https://github.com/mrkrd/thorns",
    download_url = "https://github.com/mrkrd/thorns/tarball/master",

    packages = find_packages(),
    long_description = long_description,
    classifiers = [
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python",
    ],

    platforms = ["Linux", "Windows", "FreeBSD"],
    install_requires=["numpy", "pandas"],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division, print_function, absolute_import

__author__ = "Marek Rudnicki"

import numpy as np
import pandas as pd
import warnings


def get_duration(spike_trains):
    duration = spike_trains['duration'].values[0]
    assert np.all(spike_trains['duration'] == duration)

    return duration




def psth(spike_trains, bin_size, normalize=True):

    duration = get_duration(spike_trains)
    trial_num = len(spike_trains)

    trains = spike_trains['spikes']
    all_spikes = np.concatenate(tuple(trains))

    nbins = np.ceil(duration / bin_size)

    if nbins == 0:
        return None, None

    hist, bin_edges = np.histogram(
        all_spikes,
        bins=nbins,
        range=(0, nbins*bin_size)
    )


    if normalize:
        psth = hist / bin_size / trial_num
    else:
        psth = hist


    return psth, bin_edges




def isih(spike_trains, bin_size, **kwargs):
    """Calculate inter-spike interval histogram."""

    isis = np.concatenate(
        tuple( np.diff(train) for train in spike_trains['spikes'] )
    )

    if len(isis) == 0:
        return None, None

    nbins = np.ceil(np.max(isis) / bin_size)

    hist, bin_edges = np.histogram(
        isis,
        bins=nbins,
        range=(0, nbins*bin_size),
        **kwargs
    )

    return hist, bin_edges




def entrainment(spike_trains, freq, bin_size=1e-3):
    """Calculate entrainment."""

    hist, bin_edges = isih(
        spike_trains,
        bin_size=bin_size
    )

    if hist is None:
        return np.nan


    stim_period = 1 / freq

    ent_win = (
        (bin_edges[:-1] > 0.5*stim_period)
        &
        (bin_edges[:-1] < 1.5*stim_period)
    )

    entrainment = np.sum(hist[ent_win]) / np.sum(hist)

    return entrainment



def synchronization_index(spike_trains, freq):
    """Calculate synchronization index aka vector strength."""

    if isinstance(spike_trains, pd.Series):
        spike_trains = pd.DataFrame(spike_trains).T

    all_spikes = np.concatenate( tuple(spike_trains['spikes']) )

    rate = firing_rate(spike_trains)
    if rate < 10:
        warnings.warn("Firing rate too small to reliably calculate SI: {}".format(rate))
        return np.nan


    folded = np.fmod(all_spikes, 1/freq)
    ph, edges = np.histogram(
        folded,
        bins=1000,
        range=(0, 1/freq)
    )

    # indexing trick is necessary, because the sample at 2*pi belongs
    # to the next cycle
    x = np.cos(np.linspace(0, 2*np.pi, len(ph)+1))[0:-1]
    y = np.sin(np.linspace(0, 2*np.pi, len(ph)+1))[0:-1]

    xsum2 = (np.sum(x*ph))**2
    ysum2 = (np.sum(y*ph))**2

    r = np.sqrt(xsum2 + ysum2) / np.sum(ph)

    return r


si = synchronization_index
vector_strength = synchronization_index
vs = synchronization_index


# def shuffle_spikes(spike_trains):
#     """ Get input spikes.  Randomly permute inter spikes intervals.
#     Return new spike trains.

#     """
#     new_trains = []
#     for train in spike_trains:
#         isi = np.diff(np.append(0, train)) # Append 0 in order to vary
#                                            # the onset
#         shuffle(isi)
#         shuffled_train = np.cumsum(isi)
#         new_trains.append(shuffled_train)

#     return new_trains


# def test_shuffle_spikes():
#     print "test_shuffle_spikes():"
#     spikes = [np.array([2, 3, 4]),
#               np.array([1, 3, 6])]

#     print spikes
#     print shuffle_spikes(spikes)


def firing_rate(spike_trains):
    """Calculates average firing rate."""

    if isinstance(spike_trains, pd.Series):
        spike_trains = pd.DataFrame(spike_trains).T

    duration = np.sum( spike_trains['duration'] )

    trains = spike_trains['spikes']
    spike_num = np.concatenate(tuple(trains)).size

    rate = spike_num / duration

    return rate


rate = firing_rate



def count_spikes(spike_trains):
    all_spikes = np.concatenate(tuple(spike_trains['spikes']))
    return len(all_spikes)

count = count_spikes



def correlation_index(
        spike_trains,
        coincidence_window=50e-6,
        normalize=True):
    """Compute correlation index (Joris 2006)"""

    if len(spike_trains) == 0:
        return 0

    all_spikes = np.concatenate(tuple(spike_trains['spikes']))

    if len(all_spikes) == 0:
        return 0


    Nc = 0                      # Total number of coincidences

    for spike in all_spikes:
        hits = all_spikes[
            (all_spikes >= spike) &
            (all_spikes <= spike+coincidence_window)
        ]
        Nc += len(hits) - 1



    if normalize:
        trial_num = len(spike_trains)
        firing_rate = firing_rate(spike_trains)
        duration = get_duration(spike_trains)

        norm = trial_num*(trial_num-1) * firing_rate**2 * coincidence_window * duration

        ci = Nc / norm

    else:
        ci = Nc


    return ci


ci = correlation_index


def shuffled_autocorrelogram(
        spike_trains,
        coincidence_window=50e-6,
        analysis_window=5e-3,
        normalize=True):
    """Calculate Shuffled Autocorrelogram (Joris 2006)"""

    duration = get_duration(spike_trains)
    trains = spike_trains['spikes']
    trial_num = len(trains)

    nbins = np.ceil(analysis_window / coincidence_window)

    cum = []
    for i,train in enumerate(trains):
        other_trains = list(trains)
        other_trains.pop(i)
        almost_all_spikes = np.concatenate(other_trains)

        for spike in train:
            centered = almost_all_spikes - spike
            trimmed = centered[
                (centered >= 0) & (centered < nbins*coincidence_window)
            ]
            cum.append(trimmed)

    cum = np.concatenate(cum)

    hist, bin_edges = np.histogram(
        cum,
        bins=nbins,
        range=(0, nbins*coincidence_window)
    )

    if normalize:
        firing_rate = firing_rate(spike_trains)
        norm = trial_num*(trial_num-1) * firing_rate**2 * coincidence_window * duration
        sac_half = hist / norm
    else:
        sac_half = hist

    sac = np.concatenate( (sac_half[::-1][0:-1], sac_half) )
    bin_edges = np.concatenate( (-bin_edges[::-1][1:-1], bin_edges) )


    return sac, bin_edges


sac = shuffled_autocorrelogram




def period_histogram(
        spike_trains,
        freq,
        nbins=64,               # int(spike_fs / freq)
        **kwargs
):

    all_spikes = np.concatenate( tuple(spike_trains['spikes']) )
    folded = np.fmod(all_spikes, 1/freq)
    normalized = folded * freq * 2 * np.pi

    hist, bin_edges = np.histogram(
        normalized,
        bins=nbins,
        range=(0, 2*np.pi),
        **kwargs
    )

    return hist, bin_edges
